#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <math.h>
#include <string.h>
#include "primitives.h"

int W;
int H;
unsigned char *pixels;
int window_x1;
int window_y1;
int window_x2;
int window_y2;

couleur rgb(unsigned char r, unsigned char g, unsigned char b)
{
    couleur c;
    c.r = r;
    c.g = g;
    c.b = b;
    return c;
}

couleur getpixel(int x, int y)
{
    int n = 4 * (W * y + x);
    couleur c;
    c.r = pixels[n];
    c.g = pixels[n+1];
    c.b = pixels[n+2];
    return c;
}

void putpixel(int x, int y, couleur c)
{
    int n = 4 * (W * y + x);
    if (n >= 0 && n < 4 * W *H)
    {
        pixels[n] = c.r;
        pixels[n+1] = c.g;
        pixels[n+2] = c.b;
        pixels[n+3] = 255;
    }
}


void logical2screen(double x, double y,
        int *ix, int *iy)
{
    *ix = window_x1 + x * (window_x2 - window_x1);
    *iy = window_y1 + y * (window_y2 - window_y1);
}

double delta()
{
    double delta_x = 1./(double)(1+window_x2 - window_x1);
    double delta_y = 1./(double)(1+window_y2 - window_y1);
    
    return delta_x;
}

couleur melange(double t, couleur c1, couleur c2)
{
    unsigned char r = (unsigned char) ( (1.-t) * c1.r + t * c2.r );
    unsigned char g = (unsigned char) ( (1.-t) * c1.g + t * c2.g );
    unsigned char b = (unsigned char) ( (1.-t) * c1.b + t * c2.b );
    return rgb(r,g,b);
}

void clear()
{
    memset(pixels, 255, W*H*4);
}

void initialise(int width, int height)
{
    W = width;
    H = height;
    pixels = (unsigned char *) malloc(width * height * 4);
    fenetre(0, 0, width-1, height-1);
    clear();
}

void fenetre(int x1, int y1, int x2, int y2)
{
    window_x1 = x1;
    window_x2 = x2;
    window_y1 = y1;
    window_y2 = y2;
}

void detruit()
{
    free(pixels);
}

int abs(int i)
{
    return (i<0)?-i:i;
}

void putpixel_rot_(int xc, int yc, int x, int y, couleur c)
{
    for (int i = -x; i < x; i++)
    {
        putpixel(xc+i, yc+y, c);
        putpixel(xc+i, yc-y, c);
    }
    for (int i = -y; i < y; i++)
    {
        putpixel(xc+i, yc+x, c);
        putpixel(xc+i, yc-x, c);
    }
}
 
void screen_couronne(int xc, int yc, int r1, int r2, couleur c)
{
    // algo très naif

    for (int x = xc - r1; x < xc + r1; x++)
        for (int y = yc - r1; y < yc + r1; y++)
        {
            int d2 = (x-xc) * (x-xc) + (y-yc) * (y-yc);
            if (d2 >= r2*r2 && d2 <= r1*r1)
                putpixel(x, y, c);
        }
}

void putpixel_rot(int xc, int yc, int x, int y, couleur c)
{
    putpixel(xc+x, yc+y, c);
    putpixel(xc-x, yc+y, c);
    putpixel(xc+x, yc-y, c);
    putpixel(xc-x, yc-y, c);
    putpixel(xc+y, yc+x, c);
    putpixel(xc-y, yc+x, c);
    putpixel(xc+y, yc-x, c);
    putpixel(xc-y, yc-x, c);
}
 
void screen_circle(int xc, int yc, int r, couleur c)
{
    int x = 0, y = r;
    int d = 3 - 2 * r;
    putpixel_rot(xc, yc, x, y, c);
    while (y >= x)
    {
        x++;
        if (d > 0)
        {
            y--;
            d = d + 4 * (x - y) + 10;
        }
        else
            d = d + 4 * x + 6;
        putpixel_rot(xc, yc, x, y, c);
    }
}
 
void screen_line(int x1, int y1, int x2, int y2, couleur c)
{
    // trace une ligne reliant (x1, y1) et (x2, y2)
    // avec la couleur (r,g,b) selon l'algorithme de Bresenham
    // Allez voir la page wikipédia en anglais, celle en français est nulle
    
    int dx, sx, dy, sy;
    if (x1 < x2) 
    {
        dx = x2-x1;
        sx = 1;
    }
    else
    {
        dx = x1-x2;
        sx = -1;
    }
    if (y1 < y2) 
    {
        dy = y1-y2;
        sy = 1;
    }
    else
    {
        dy = y2-y1;
        sy = -1;
    }

    int err = dx + dy;

    while(true)
    {
        putpixel(x1, y1, c);
        if (x1 == x2 && y1 == y2)
            break;
        int e2 = 2*err;
        if (e2 >= dy)
        {
            err += dy;
            x1 += sx;
        }
        if (e2 <= dx)
        {
            err += dx;
            y1 += sy;
        }
    }
}

void ligne(double x1, double y1, double x2, double y2, couleur c)
{
    int ix1, iy1, ix2, iy2;
    logical2screen(x1, y1, &ix1, &iy1);
    logical2screen(x2, y2, &ix2, &iy2);

    screen_line(ix1, iy1, ix2, iy2, c);
}

void cercle_plein(double xc, double yc, double r, couleur c)
{
    int ixc, iyc, ir;
    logical2screen(xc, yc, &ixc, &iyc);
    ir = r * (window_x2 - window_x1);

    screen_couronne(ixc, iyc, ir, 0, c);
}

void cercle(double xc, double yc, double r, couleur c)
{
    int ixc, iyc, ir;
    logical2screen(xc, yc, &ixc, &iyc);
    ir = r * (window_x2 - window_x1);

    screen_circle(ixc, iyc, ir, c);
}

void couronne(double xc, double yc, double r, double e, couleur c)
{
    int ixc, iyc, ir, ii;
    logical2screen(xc, yc, &ixc, &iyc);
    ir = r * (window_x2 - window_x1);
    ii = (r-e) * (window_x2 - window_x1);

    screen_couronne(ixc, iyc, ir, ii, c);
}

unsigned char* get_pixels()
{
    return pixels;
}

bool eq_couleur(couleur a, couleur b)
{
    return a.r == b.r && a.g == b.g && a.b == b.b;
}

unsigned char add_clamp(unsigned char a, unsigned char b)
{
    int ia = 255-a;
    int ib = 255-b;
    int s = ia+ib;
    if (s > 255) s = 255;
    return 255 - s;
}

couleur couleur_multiply(couleur a, couleur b)
{
    couleur m;

    // 255 * (r/255 * r'/255) = (rr')/255
    m.r = add_clamp(a.r, b.r);
    m.g =  add_clamp(a.g, b.g);
    m.b =  add_clamp(a.b, b.b);
    return m;
}

void flood_fill_bord(int x, int y, couleur new, couleur bord)
{
    int w = 1 + window_x2 - window_x1;
    int h = 1 + window_y2 - window_y1;

    // Prévient les cas dégénérés dus à des arrondis
    if (x < 0 || x > W-1 || y < 0 || y > H-1) return;
    if (x < window_x1 || x > window_x2) return;
    if (y < window_y1 || y > window_y2) return;

    int stacked = 0;
    int limit = w * h * 4 - 1;
    int *stack_x = (int *) malloc(sizeof(int) * limit);
    int *stack_y = (int *) malloc(sizeof(int) * limit);
    char *visited = (char *) malloc(sizeof(char) * w * h);
    memset(visited, 0, sizeof(char) * w * h);

    couleur old = getpixel(x, y);

    stack_x[stacked] = x;
    stack_y[stacked] = y;
    stacked++;

    while (stacked != 0)
    {
        stacked--;
        int x = stack_x[stacked];
        int y = stack_y[stacked];

        int xl = x - window_x1;
        int yl = y - window_y1;

        if (visited[yl*w+xl] == 2) continue;
        if (x < 0 || x > W-1 || y < 0 || y > H-1) continue;
        if (x < window_x1 || x > window_x2) continue;
        if (y < window_y1 || y > window_y2) continue;

        if (eq_couleur(getpixel(x, y), bord)) continue;
        couleur c = getpixel(x, y);
        putpixel(x, y, couleur_multiply(c, new));
        visited[yl * w + xl] = 2;
     
        if (x > window_x1 && visited[yl * w + xl - 1] == 0) {
            visited[yl * w + xl - 1] = 1;
            stack_x[stacked] = x-1;
            stack_y[stacked] = y;
            stacked++;
        }
        if (x < window_x2 && visited[yl * w + xl + 1] == 0) {
            visited[yl * w + xl + 1] = 1;
            stack_x[stacked] = x+1;
            stack_y[stacked] = y;
            stacked++;
        }
        if (y > window_y1 && visited[(yl-1) * w + xl] == 0) {
            visited[(yl-1) * w + xl] = 1;
            stack_x[stacked] = x;
            stack_y[stacked] = y-1;
            stacked++;
        }
        if (y < window_y2 && visited[(yl+1) * w + xl] == 0) {
            visited[(yl+1) * w + xl] = 1;
            stack_x[stacked] = x;
            stack_y[stacked] = y+1;
            stacked++;
        }
    }

    free(stack_x);
    free(stack_y);
    free(visited);
}

void flood_fill(int x, int y, couleur new)
{
    int w = 1 + window_x2 - window_x1;
    int h = 1 + window_y2 - window_y1;

    // Prévient les cas dégénérés dus à des arrondis
    if (x < 0 || x > W-1 || y < 0 || y > H-1) return;
    if (x < window_x1 || x > window_x2) return;
    if (y < window_y1 || y > window_y2) return;

    int stacked = 0;
    int limit = w * h * 4 - 1;
    int *stack_x = (int *) malloc(sizeof(int) * limit);
    int *stack_y = (int *) malloc(sizeof(int) * limit);
    char *visited = (char *) malloc(sizeof(char) * w * h);
    memset(visited, 0, sizeof(char) * w * h);

    couleur old = getpixel(x, y);

    stack_x[stacked] = x;
    stack_y[stacked] = y;
    stacked++;

    while (stacked != 0)
    {
        stacked--;
        int x = stack_x[stacked];
        int y = stack_y[stacked];

        int xl = x - window_x1;
        int yl = y - window_y1;

        if (visited[yl*w+xl] == 2) continue;
        if (x < 0 || x > W-1 || y < 0 || y > H-1) continue;
        if (x < window_x1 || x > window_x2) continue;
        if (y < window_y1 || y > window_y2) continue;

        if (!eq_couleur(getpixel(x, y), old)) continue;
        putpixel(x, y, new);
        visited[yl * w + xl] = 2;
     
        if (x > window_x1 && visited[yl * w + xl - 1] == 0) {
            visited[yl * w + xl - 1] = 1;
            stack_x[stacked] = x-1;
            stack_y[stacked] = y;
            stacked++;
        }
        if (x < window_x2 && visited[yl * w + xl + 1] == 0) {
            visited[yl * w + xl + 1] = 1;
            stack_x[stacked] = x+1;
            stack_y[stacked] = y;
            stacked++;
        }
        if (y > window_y1 && visited[(yl-1) * w + xl] == 0) {
            visited[(yl-1) * w + xl] = 1;
            stack_x[stacked] = x;
            stack_y[stacked] = y-1;
            stacked++;
        }
        if (y < window_y2 && visited[(yl+1) * w + xl] == 0) {
            visited[(yl+1) * w + xl] = 1;
            stack_x[stacked] = x;
            stack_y[stacked] = y+1;
            stacked++;
        }
    }

    free(stack_x);
    free(stack_y);
    free(visited);
}

void remplit(double x, double y, couleur c)
{
    int ix, iy;
    logical2screen(x, y, &ix, &iy);
    flood_fill(ix, iy, c);
}


void remplit_jusquau_bord(double x, double y, couleur c, couleur bord)
{
    int ix, iy;
    logical2screen(x, y, &ix, &iy);
    flood_fill_bord(ix, iy, c, bord);
}

void efface()
{
    memset(pixels, 0, W * H * 4);
}
